--[[
    Aaron MacDonald
    November 25, 2010
]]

require('graphics')
require('colour_cycle')
require('utility')

require('starfield')

----------------------------------------------------------------------

local FPS = 60
local SECONDS_PER_FRAME = 1 / FPS

local accumulatedTime

local isPaused

local KEY_QUIT = 'escape'
local KEY_PAUSE = 'p'

----------------------------------------------------------------------

local KEY_UP = 'w'
local KEY_LEFT = 'a'
local KEY_DOWN = 's'
local KEY_RIGHT = 'd'

----------------------------------------------------------------------

local MOUSE_SPRITE_NAME = 'target'
local MOUSE_SPRITE
local MOUSE_COLOUR_CYCLE

local MOUSE_COLOURS = { {255, 255, 255},
                        {255, 225,  64} }
local MOUSE_COLOUR_SPEED = 0.1

----------------------------------------------------------------------

local ARENA_WIDTH
local ARENA_HEIGHT

local PLAYER_SPRITE_NAME = 'sphinx'
local SPARKLE_SPRITE_NAME = 'sparkle'

-- pixels/frame
local PLAYER_TOP_SPEED = 8
local PLAYER_ACCEL = 0.25

local PLAYER_ROT_SPEED = 0.2

local PLAYER_GUN_RANGE = 600

local NUM_GUNSIGHT_SPARKLES = 25

local SPARKLE_FADE_STEP = 0.05

local SPARKLE_START_COLOUR = {128, 191, 255}
local SPARKLE_END_COLOUR = {255, 255, 255}

local sparkleSprite
local sparkles

local playerSprite

local NUM_ENGINE_PARTICLES = 1000
local ENGINE_EMISSION_RATE = 300
local ENGINE_SPEED_MIN = 150
local ENGINE_SPEED_MAX = 250
local ENGINE_SIZE_MIN = 0.25
local ENGINE_SIZE_MAX = 1.5
local ENGINE_START_COLOUR = {255, 255, 255, 255}
local ENGINE_END_COLOUR = {255, 223, 0, 0}
local ENGINE_PARTICLE_LIFE_MIN = 0.1
local ENGINE_PARTICLE_LIFE_MAX = 0.75
local ENGINE_SPREAD = math.pi / 6

local playerEngine

local playerX
local playerY

local playerDX
local playerDY

local playerAngle

----------------------------------------------------------------------

function love.load()
    math.randomseed( os.time() )

    accumulatedTime = 0
    isPaused = false

    ----------------------------

    starfield.init()

    ----------------------------

    love.mouse.setVisible(false)

    MOUSE_SPRITE = graphics.createSprite(MOUSE_SPRITE_NAME)

    MOUSE_COLOUR_CYCLE = colourCycle.createCycle(MOUSE_COLOURS,
      MOUSE_COLOUR_SPEED)

    ----------------------------

    ARENA_WIDTH = love.graphics.getWidth()
    ARENA_HEIGHT = love.graphics.getHeight()

    playerSprite = graphics.createSprite(PLAYER_SPRITE_NAME)
    sparkleSprite = graphics.createSprite(SPARKLE_SPRITE_NAME)

    sparkles = {}
    for i = 1, NUM_GUNSIGHT_SPARKLES do
        sparkles[i] = {
            range = math.random() * PLAYER_GUN_RANGE,
            fade = math.random() * 2
        }
    end

    playerX = ARENA_WIDTH / 2
    playerY = ARENA_HEIGHT / 2

    playerDX = 0
    playerDY = 0

    playerAngle = 0

    local emitter = love.graphics.newParticleSystem(
      sparkleSprite.drawable, NUM_ENGINE_PARTICLES)
    emitter:setEmissionRate(ENGINE_EMISSION_RATE)
    emitter:setSpeed(ENGINE_SPEED_MIN, ENGINE_SPEED_MAX)
    emitter:setSize(ENGINE_SIZE_MIN, ENGINE_SIZE_MAX)
    emitter:setColor(
      ENGINE_START_COLOUR[1],
      ENGINE_START_COLOUR[2],
      ENGINE_START_COLOUR[3],
      ENGINE_START_COLOUR[4],
      ENGINE_END_COLOUR[1],
      ENGINE_END_COLOUR[2],
      ENGINE_END_COLOUR[3],
      ENGINE_END_COLOUR[4]
    )
    emitter:setLifetime(-1)
    emitter:setParticleLife(ENGINE_PARTICLE_LIFE_MIN,
      ENGINE_PARTICLE_LIFE_MAX)
    emitter:setSpread(ENGINE_SPREAD)

    playerEngine = graphics.createParticleSprite(emitter,
      sparkleSprite.drawable, ENGINE_PARTICLE_LIFE_MAX,
      ENGINE_SPEED_MAX, ENGINE_SIZE_MAX)
end

----------------------------------------------------------------------

function love.keypressed( key, unicode )
    if key == KEY_QUIT then
        love.event.push('q')
    elseif key == KEY_PAUSE then
        isPaused = not isPaused
    end
end

----------------------------------------------------------------------

function love.focus(hasFocus)
    if not hasFocus then
        isPaused = true
    end
end

----------------------------------------------------------------------

local function accelerate(val, dir)
    local newVal = val

    if (dir == 0) and (val ~= 0) then
        -- Slow to a stop
        local sign = utility.sign(val)
        newVal = math.abs(val) - PLAYER_ACCEL
        if newVal < 0 then
            newVal = 0
        else
            newVal = newVal * sign
        end
    elseif (dir ~= 0) and (val ~= (PLAYER_TOP_SPEED * dir)) then
        newVal = val + (dir * PLAYER_ACCEL)
        if math.abs(newVal) > math.abs(PLAYER_TOP_SPEED * dir) then
            newVal = PLAYER_TOP_SPEED * dir
        end
    end

    return newVal
end

local function doUpdate()
    starfield.update()
    colourCycle.update(MOUSE_COLOUR_CYCLE)

    -- Determine thrust

    local dx = 0
    local dy = 0

    if love.keyboard.isDown(KEY_LEFT) then
        dx = dx - 1
    end
    if love.keyboard.isDown(KEY_RIGHT) then
        dx = dx + 1
    end

    if love.keyboard.isDown(KEY_UP) then
        dy = dy - 1
    end
    if love.keyboard.isDown(KEY_DOWN) then
        dy = dy + 1
    end

    if (dx ~= 0) or (dy ~= 0) then
        playerEngine.drawable:setDirection(math.atan2(-dy, -dx))
        if not playerEngine.drawable:isActive() then
            playerEngine.drawable:start()
        end
    else
        playerEngine.drawable:stop()
    end

    playerDX = accelerate(playerDX, dx)
    playerDY = accelerate(playerDY, dy)

    -- Update position

    playerX = (playerX + playerDX) % ARENA_WIDTH
    if playerX < 0 then
        playerX = playerX + ARENA_WIDTH
    end

    playerY = (playerY + playerDY) % ARENA_HEIGHT
    if playerY < 0 then
        playerY = playerY + ARENA_HEIGHT
    end

    -- Angle

    local mx = love.mouse.getX()
    local my = love.mouse.getY()

    playerAngle = math.atan2(my - playerY, mx - playerX)

    -- Sparkles
    for _, s in ipairs(sparkles) do
        s.fade = s.fade - SPARKLE_FADE_STEP
        if s.fade < 0 then
            s.range = math.random() * PLAYER_GUN_RANGE
            s.fade = s.fade + 2
        end
    end

    -- Engines

    playerEngine.drawable:update(SECONDS_PER_FRAME)
end

function love.update(dt)
    accumulatedTime = accumulatedTime + dt
    while accumulatedTime >= SECONDS_PER_FRAME do
        if not isPaused then
            doUpdate()
        end
        accumulatedTime = accumulatedTime - SECONDS_PER_FRAME
    end
end

----------------------------------------------------------------------

function love.draw()
    starfield.draw()

    graphics.drawWrappedSprite(playerEngine, playerX, playerY,
      ARENA_WIDTH, ARENA_HEIGHT)

    --love.graphics.draw(playerEngine, playerX, playerY)

    -- Gunsight

    for _, s in ipairs(sparkles) do
        local x = (s.range * math.cos(playerAngle)) + playerX
        local y = (s.range * math.sin(playerAngle)) + playerY

        local fade
        if s.fade < 1 then
            fade = s.fade
        else
            fade = 1 - s.fade
        end

        local sparkleColour = graphics.interpolatedColour(
          SPARKLE_END_COLOUR, SPARKLE_START_COLOUR, fade)

        love.graphics.setColor(sparkleColour[1], sparkleColour[2],
          sparkleColour[3], 255 * fade)

        graphics.drawWrappedSprite(sparkleSprite, x, y,
          ARENA_WIDTH, ARENA_HEIGHT)
    end

    graphics.setColourWhite()

    -- Player

    graphics.drawWrappedSprite(playerSprite, playerX, playerY,
      ARENA_WIDTH, ARENA_HEIGHT, playerAngle)

    -- Mouse

    colourCycle.setColour(MOUSE_COLOUR_CYCLE)

    graphics.drawSprite(MOUSE_SPRITE,
      love.mouse.getX(), love.mouse.getY())

    graphics.setColourWhite()
end